#
# Copyright 2020, Xiaomi.
# All rights reserved.
# Author: huyumei@xiaomi.com
# 

# read properties for client / consumer / producer

from talos.client.TalosErrors import InvalidArgumentError
from talos.client.Constants import Constants
from talos.thrift.message.ttypes import MessageOffset
from talos.thrift.common.ttypes import GalaxyTalosException
from talos.thrift.common.ttypes import ErrorCode
import functools
import time
import re
import uuid
import threading
import socket
import math


def synchronized(func):
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        with self.synchronizedLock:
            return func(self, *args, **kwargs)
    return wrapper


class ReadWriteLock:
    """ A lock object that allows many simultaneous "read locks", but
    only one "write lock." """
    def __init__(self):
        self.rlock = threading.Lock()
        self.wlock = threading.Lock()
        self.reader = 0

    def write_acquire(self):
        self.wlock.acquire()

    def write_release(self):
        self.wlock.release()

    def read_acquire(self):
        self.rlock.acquire()
        self.reader += 1
        if self.reader == 1:
            self.wlock.acquire()
        self.rlock.release()

    def read_release(self):
        self.rlock.acquire()
        self.reader -= 1
        if self.reader == 0:
            self.wlock.release()
        self.rlock.release()


def current_time_mills():
    return int(time.time() * 1000)


def update_message(message=None, messageType=None):
    if not message.createTimestamp:
        message.createTimestamp = current_time_mills()

    if not message.messageType:
        message.messageType = messageType


def check_message_list_validity(msgList=None):
    totalSize = 0
    for message in msgList:
        check_message_validity(message)
        totalSize += len(message.message)

    # is too large, and not just throw exception;
    if totalSize > Constants.TALOS_SINGLE_MESSAGE_BYTES_MAXIMAL * 2:
        raise InvalidArgumentError("Total Messages byte must less than " +
                                   str(Constants.TALOS_SINGLE_MESSAGE_BYTES_MAXIMAL * 2))


def check_message_validity(message=None):
    check_message_len_validity(message)
    check_message_sequence_number_validity(message)
    check_message_type_validity(message)


def check_message_len_validity(message=None):
    if not message.message:
        raise InvalidArgumentError("Field \"message\" must be set")

    data = message.message
    if len(data) > Constants.TALOS_SINGLE_MESSAGE_BYTES_MAXIMAL\
            or len(data) < Constants.TALOS_SINGLE_MESSAGE_BYTES_MINIMAL:
        raise InvalidArgumentError("Data must be less than or equal to " +
                                   str(Constants.TALOS_SINGLE_MESSAGE_BYTES_MAXIMAL) +
                                   " bytes, got bytes: " + str(len(data)))


def check_message_sequence_number_validity(message=None):
    if not message.sequenceNumber:
        return

    sequenceNumber = message.sequenceNumber
    if len(sequenceNumber) > Constants.TALOS_PARTITION_KEY_LENGTH_MAXIMAL\
            or len(sequenceNumber) < Constants.TALOS_PARTITION_KEY_LENGTH_MINIMAL:
        raise InvalidArgumentError("Invalid sequenceNumber which length must be at least "
                                   + str(Constants.TALOS_PARTITION_KEY_LENGTH_MINIMAL) +
                                   " and at most " + str(Constants.TALOS_PARTITION_KEY_LENGTH_MAXIMAL)
                                   + ", got " + str(len(sequenceNumber)))


def check_message_type_validity(message=None):
    if not message.messageType:
        raise InvalidArgumentError("Filed \"messageType\" must be set")


def check_parameter_range(parameter=None, value=None, minValue=None, maxValue=None):
    if value < minValue or value > maxValue:
        raise InvalidArgumentError(parameter + " should be in range [" + str(minValue) +
                                   ", " + str(maxValue) + "], got: " + str(value))


def generate_request_sequence_id(clientId=None, requestId=None):
    check_name_validity(clientId)
    return clientId + Constants.TALOS_IDENTIFIER_DELIMITER + str(requestId.get_and_set(requestId.value + 1))


# The format of cloud topicName is: orgId/topicName
def check_cloud_topic_name_validity(topicName=None):
    if not topicName or len(topicName) == 0:
        raise InvalidArgumentError("Got null topicName")
    items = topicName.split(Constants.TALOS_CLOUD_TOPIC_NAME_DELIMITER)
    # either 'xxx/xxx/'(split 2), '/xxx'(split 2) or 'xx//xx'(split 3) are invalid
    if len(items) != 2 or topicName.endswith(Constants.TALOS_CLOUD_TOPIC_NAME_DELIMITER) or not topicName.startswith(Constants.TALOS_CLOUD_ORG_PREFIX):
        raise InvalidArgumentError("The format of topicName used by cloud-manager must be: orgId/topicName")
    # check real topic name validity
    check_name_validity(items[1])


def check_name_validity(name=None):
    if (not name) or len(name) <= 0:
        return
    if (not re.match(Constants.TALOS_NAME_REGEX, name)) or len(name) > 80:
        raise InvalidArgumentError("invalid str: " + name + ". please name the str only " +
                                   "with the regex set: [a-zA-Z0-9_-]. Its length must be" +
                                   " [1, 80] and cannot start with '_' or '-'.")


def generate_client_id(clientIp=None, prefix=None):
    check_name_validity(prefix)
    return clientIp.replace(".", "-") + "-" + prefix + generate_client_id_()


def generate_client_id_():
    return str(current_time_mills()) + str(uuid.uuid4())[0:8]


def check_topic_name(topicName=None):
    if Constants.TALOS_CLOUD_TOPIC_NAME_DELIMITER in topicName:
        raise InvalidArgumentError("The topic name format in TopicAndPartition should" +
                                   " not be: orgId/topicName")


def check_start_offset_validity(startOffset=None):
    if startOffset >= 0 or startOffset == MessageOffset.START_OFFSET or \
            startOffset == MessageOffset.LATEST_OFFSET:
        return
    raise InvalidArgumentError("invalid startOffset: " + str(startOffset) +
                               ". It must be greater than or equal to 0, " +
                               "or equal to MessageOffset.START_OFFSET/MessageOffset.LATEST_OFFSET")


def get_error_code(throwable=None):
    if isinstance(throwable, GalaxyTalosException):
        return throwable.errorCode
    else:
        pass


def is_topic_not_exist(throwable=None):
    return get_error_code(throwable) == ErrorCode.TOPIC_NOT_EXIST


def is_partition_not_serving(throwable=None):
    return get_error_code(throwable) == ErrorCode.PARTITION_NOT_SERVING


def is_partition_not_exist(throwable=None):
    return get_error_code(throwable) == ErrorCode.PARTITION_NOT_EXIST


def is_offset_out_of_range(throwable=None):
    return get_error_code(throwable) == ErrorCode.MESSAGE_OFFSET_OUT_OF_RANGE


def get_topic_name_by_resource_name(topicTalosResourceName=None):
    itemList = topicTalosResourceName.split(Constants.TALOS_IDENTIFIER_DELIMITER)
    assert len(itemList) >= 3
    return itemList[len(itemList) - 2]


def get_put_msg_failed_delay(retry=None, producerConfig=None):
    delayTimes = int(math.pow(2, retry))
    return min(delayTimes * producerConfig.get_put_message_base_backoff_time(),
               producerConfig.get_put_message_max_backoff_time())


def sleep_pause_time(pauseTime=None):
    if pauseTime > 0:
        try:
            time.sleep(pauseTime)
        except Exception as ie:
            raise RuntimeError("thread sleep failed" + ie.message)


class HashCode:

    def convert_n_bytes(self, n, b):
        bits = b * 8
        return (n + 2 ** (bits - 1)) % 2 ** bits - 2 ** (bits - 1)

    def convert_4_bytes(self, n):
        return self.convert_n_bytes(n, 4)

    @classmethod
    def HashCode(cls, s):
        h = 0
        n = len(s)
        for i, c in enumerate(s):
            h = h + ord(c) * 31 ** (n - 1 - i)
        return cls().convert_4_bytes(h)


class SerializeFormat(object):

    format_i16 = '>h'
    format_i32 = '>i'
    format_i64 = '>Q'


def get_host_ip():
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        s.connect(('8.8.8.8', 80))
        ip = s.getsockname()[0]
    finally:
        s.close()
    return ip




#
# Copyright 2020, Xiaomi.
# All rights reserved.
# Author: huyumei@xiaomi.com
#

import unittest

from talos.thrift.message.ttypes import MessageAndOffset
from talos.thrift.message.ttypes import Message
from talos.thrift.consumer.ttypes import UpdateOffsetResponse
from talos.thrift.consumer.ttypes import QueryOffsetResponse
from talos.thrift.topic.ttypes import TopicTalosResourceName
from talos.thrift.topic.ttypes import TopicAndPartition
from talos.consumer.SimpleConsumer import SimpleConsumer
from talos.consumer.TalosMessageReader import TalosMessageReader
from talos.consumer.MessageProcessor import MessageProcessor
from talos.client.TalosClientFactory import ConsumerClient
from talos.client.TalosClientConfig import TalosClientConfig
from talos.client.TalosClientConfigkeys import TalosClientConfigKeys
from mock import Mock


class TestMessageProcessor(MessageProcessor):
    topicAndPartition = TopicAndPartition
    messageOffset = long

    def init(self, topicAndPartition=None, messageOffset=None):
        self.topicAndPartition = topicAndPartition
        self.messageOffset = messageOffset

    def process(self, messages=None, messageCheckpointer=None):
        assert messageCheckpointer.checkpoint()
        # checkpoint low messageOffset
        assert not messageCheckpointer.checkpoint(self.messageOffset)
        # checkpoint high messageOffset
        assert not messageCheckpointer.checkpoint(messages.get(messages.size()-1).getMessageOffset())

    def shutdown(self, messageCheckpointer=None):
        assert not messageCheckpointer.checkpoint()

    def get_topic_and_partition(self):
        return self.topicAndPartition

    def get_message_offset(self):
        return self.messageOffset


class test_TalosMessageReader(unittest.TestCase):
    properties = dict()
    consumerConfig = TalosClientConfig
    topicName = "MyTopic"
    resourceName = "12345#MyTopic#34595fkdiso456i390"
    partitionId = 7
    consumerGroup = "MyConsumerGroup"
    workerId = "workerId"
    topicAndPartition = TopicAndPartition(topicName, TopicTalosResourceName(resourceName), partitionId)
    startMessageOffset = 1

    messageAndOffsetList = []
    messageAndOffsetList2 = []
    consumerClientMock = ConsumerClient
    simpleConsumerMock = SimpleConsumer

    messageProcessor = TestMessageProcessor()
    messageProcessorMock = TestMessageProcessor()
    messageReader = TalosMessageReader()

    def setUp(self):
        messageAndOffset1 = MessageAndOffset(Message("message1"), self.startMessageOffset)
        messageAndOffset2 = MessageAndOffset(Message("message2"), self.startMessageOffset + 1)
        messageAndOffset3 = MessageAndOffset(Message("message3"), self.startMessageOffset + 2)
        self.messageAndOffsetList = [messageAndOffset1, messageAndOffset2, messageAndOffset3]
        messageAndOffset4 = MessageAndOffset(Message("message4"), self.startMessageOffset + 3)
        messageAndOffset5 = MessageAndOffset(Message("message5"), self.startMessageOffset + 4)
        self.messageAndOffsetList2 = [messageAndOffset4, messageAndOffset5]

        self.consumerClientMock = ConsumerClient

    def test_DefaultCheckpointOffset(self):
        talosConfigKeys = TalosClientConfigKeys()
        self.properties[talosConfigKeys.GALAXY_TALOS_CONSUMER_CHECKPOINT_AUTO_COMMIT] = True
        self.properties[talosConfigKeys.GALAXY_TALOS_CONSUMER_FETCH_INTERVAL] = 0
        self.properties[talosConfigKeys.GALAXY_TALOS_CONSUMER_COMMIT_OFFSET_INTERVAL] = 0
        self.properties[talosConfigKeys.GALAXY_TALOS_CONSUMER_COMMIT_OFFSET_THRESHOLD] = 0
        self.properties[TalosClientConfigKeys.GALAXY_TALOS_SERVICE_ENDPOINT] = "testURI"
        self.consumerConfig = TalosClientConfig(self.properties)

        self.messageProcessorMock = TestMessageProcessor()
        messageReader = TalosMessageReader(consumerConfig=self.consumerConfig)
        messageReader.set_simple_consumer(self.simpleConsumerMock)
        messageReader.set_consumer_client(self.consumerClientMock)
        messageReader.set_consumer_group(self.consumerGroup)
        messageReader.set_message_processor(self.messageProcessorMock)
        messageReader.set_topic_and_partition(self.topicAndPartition)
        messageReader.set_worker_id(self.workerId)

        self.messageProcessorMock.init(topicAndPartition=self.topicAndPartition,
                                       messageOffset=self.startMessageOffset)
        queryOffsetResponse = QueryOffsetResponse(self.startMessageOffset - 1)
        self.consumerClientMock.query_offset = Mock(return_value=queryOffsetResponse)
        self.simpleConsumerMock.fetch_message = Mock(side_effect=[self.messageAndOffsetList, self.messageAndOffsetList2])
        self.messageProcessorMock.process = Mock(return_value=None)
        updateOffsetResponse = UpdateOffsetResponse(True)
        self.consumerClientMock.update_offset = Mock(return_value=updateOffsetResponse)
        self.messageProcessorMock.shutdown = Mock(return_value=None)

        messageReader.init_start_offset()
        messageReader.fetch_data()
        messageReader.fetch_data()
        messageReader.commit_check_point()

    def test_UserCheckpointOffset(self):
        self.properties[TalosClientConfigKeys.GALAXY_TALOS_CONSUMER_CHECKPOINT_AUTO_COMMIT] = False
        self.properties[TalosClientConfigKeys.GALAXY_TALOS_CONSUMER_FETCH_INTERVAL] = 0
        self.properties[TalosClientConfigKeys.GALAXY_TALOS_SERVICE_ENDPOINT] = "testURI"
        consumerConfig = TalosClientConfig(self.properties)

        messageReader = TalosMessageReader(consumerConfig)
        messageReader.set_simple_consumer(self.simpleConsumerMock)
        messageReader.set_consumer_client(self.consumerClientMock)
        messageReader.set_consumer_group(self.consumerGroup)
        messageReader.set_message_processor(self.messageProcessor)
        messageReader.set_topic_and_partition(self.topicAndPartition)
        messageReader.set_worker_id(self.workerId)

        queryOffsetResponse = QueryOffsetResponse(self.startMessageOffset - 1)
        self.consumerClientMock.query_offset = Mock(return_value=queryOffsetResponse)
        messageReader.init_start_offset()
        self.assertEquals(self.messageProcessor.get_message_offset(), self.startMessageOffset)
        self.assertEquals(self.messageProcessor.get_topic_and_partition(), self.topicAndPartition)

        self.simpleConsumerMock.fetch_message = self.messageAndOffsetList
        updateOffsetResponse = UpdateOffsetResponse(True)
        self.consumerClientMock.update_offset = Mock(return_value=updateOffsetResponse)
        self.simpleConsumerMock.fetch_message = self.messageAndOffsetList2

        messageReader.fetch_data()
        messageReader.fetch_data()
        messageReader.commit_check_point()


if __name__ == '__main__':
    unittest.main()



#
# Copyright 2020, Xiaomi.
# All rights reserved.
# Author: huyumei@xiaomi.com
# 

import abc


class UserMessageCallback:
    #
    # Implement this method to process messages that successfully put to server
    # user can get 'messageList', 'partitionId' and 'isSuccessful' by userMessageResult
    #
    @abc.abstractmethod
    def on_success(self, userMessageResult=None):
        pass

    #
    # Implement this method to process messages failed to put to server
    # user can get 'messageList', 'partitionId' and 'cause' by userMessageResult
    #
    @abc.abstractmethod
    def on_error(self, userMessageResult=None):
        pass




#
# Copyright 2020, Xiaomi.
# All rights reserved.
# Author: huyumei@xiaomi.com
# 
 

class UserMessageResult:
    messageList = []
    topicName = str
    partitionId = int
    successful = bool
    cause = Exception

    def __init__(self, messageList=None, topicName=None, partitionId=None):
        self.messageList = messageList
        self.topicName = topicName
        self.partitionId = partitionId
        self.successful = False
        self.cause = None

    def get_topic_name(self):
        return self.topicName

    def get_partition_id(self):
        return self.partitionId

    def is_successful(self):
        return self.successful

    def get_message_list(self):
        return self.messageList

    def get_cause(self):
        return self.cause

    def set_successful(self, successful=None):
        self.successful = successful

    def set_cause(self, cause=None):
        self.cause = cause



#
# Copyright 2020, Xiaomi.
# All rights reserved.
# Author: huyumei@xiaomi.com
#

from talos.consumer.MessageReader import MessageReader
from talos.thrift.consumer.ttypes import QueryOffsetRequest
from talos.thrift.consumer.ttypes import CheckPoint
from talos.thrift.consumer.ttypes import UpdateOffsetRequest
from talos.thrift.message.ttypes import GetPartitionOffsetRequest
from talos.consumer.MessageCheckpointer import MessageCheckpointer
from talos.utils import Utils
from talos.thrift.common.ttypes import GalaxyTalosException
import traceback
import logging
import time


class TalosMessageReader(MessageReader, MessageCheckpointer):
    logger = logging.getLogger("TalosMessageReader")

    def __init__(self, consumerConfig=None):
        MessageReader.__init__(self, consumerConfig)

    def init_start_offset(self):
        # get last commit offset or init by outer checkPoint
        readingStartOffset = int
        # when consumer  starting up, checking:
        # 1) whether not exist last commit offset, which means 'readingStartOffset==-1'
        # 2) whether reset offset
        # 3) note that: the priority of 'reset-config' is larger than 'outer-checkPoint'

        # there are following situations
        # 1) reset Offset When Start
        #    1) -1 ———— > set startOffset to - 1 and reading from start
        #    2) -2 ———— > set startOffset to endOffset and reading from end
        # 2) not reset Offset When Start
        #    1) outerCheckPoint is set ———— > reading from outerCheckPoint
        #    2) check point is not exist ———— > same with "1) reset Offset When Start"
        #    3) check point is exist ———— > reading from check point

        if self.consumerConfig.is_reset_offset_when_start():
            readingStartOffset = self.reset_read_start_offset(
                self.consumerConfig.get_reset_offset_value_when_start())
        elif (not self.outerCheckPoint) and self.outerCheckPoint >= 0:
            readingStartOffset = self.outerCheckPoint
            # burn after reading the first time
            self.outerCheckPoint = None
        else:
            queryStartOffset = self.query_start_offset()
            if queryStartOffset == -1:
                readingStartOffset = self.reset_read_start_offset(
                    self.consumerConfig.get_reset_offset_value_when_start())
            else:
                readingStartOffset = queryStartOffset

        self.startOffset.get_and_set(readingStartOffset)
        # guarantee lastCommitOffset and finishedOffset correct
        if self.startOffset.value > 0:
            self.lastCommitOffset = self.finishedOffset = self.startOffset.value - 1
        self.logger.info("Init startOffset: " + str(self.startOffset)
                         + " lastCommitOffset: " + str(self.lastCommitOffset)
                         + " for partition: " + self.topicAndPartition.topicName
                         + str(self.topicAndPartition.partitionId))
        self.messageProcessor.init(self.topicAndPartition, self.startOffset.value)

    def reset_read_start_offset(self, resetOffsetValueWhenStart=None):
        # if resetOffsetValueWhenStart is -1 , just set startOffset to -1 and would read from the start
        # if resetOffsetValueWhenStart is -2 , we should get the endOffset from the server for reading from the end
        if resetOffsetValueWhenStart == -1:
            return -1

        request = GetPartitionOffsetRequest(self.topicAndPartition)
        response = self.messageClient.get_partition_offset(request)
        return response.endOffset + 1

    def commit_check_point(self):
        self.inner_checkpoint()
        self.messageProcessor.shutdown(self)

    def fetch_data(self):
        # control fetch qps
        currentTime = Utils.current_time_mills()
        # sleep time unit is seconds
        if currentTime - self.lastFetchTime < self.fetchInterval:
            time.sleep((self.lastFetchTime + self.fetchInterval - currentTime) / 1000)

        # fetch data and process them
        try:
            if self.logger.isEnabledFor:
                self.logger.debug("Reading message from offset: " + str(self.startOffset.value) +
                                  " of partition: " + str(self.topicAndPartition.partitionId))

            messageList = self.simpleConsumer.fetch_message(startOffset=self.startOffset.value,
                                                            maxFetchedNumber=self.consumerConfig.get_max_fetch_msg_num())
            self.lastFetchTime = Utils.current_time_mills()

            # return when no message got
            if not messageList or len(messageList) == 0:
                self.check_and_commit(False)
                return

            #
            # Note: We guarantee the committed offset must be the messages that
            # have been processed by user 's MessageProcessor;
            self.finishedOffset = messageList[len(messageList) - 1].messageOffset
            self.messageProcessor.process(messageList, self)
            self.startOffset.get_and_set(self.finishedOffset + 1)
            self.check_and_commit(True)
        except Exception as e:
            if isinstance(e, GalaxyTalosException):
                self.logger.error(e.errMsg)
            self.logger.error("Error when getting messages from topic: "
                              + self.topicAndPartition.topicTalosResourceName.topicTalosResourceName
                              + " partition: " + str(self.topicAndPartition.partitionId)
                              + str(traceback.format_exc()))
            self.process_fetch_exception(e)
            self.lastFetchTime = Utils.current_time_mills()

    def query_start_offset(self):
        queryOffsetRequest = QueryOffsetRequest(self.consumerGroup, self.topicAndPartition)
        queryOffsetResponse = self.consumerClient.query_offset(queryOffsetRequest)

        committedOffset = queryOffsetResponse.msgOffset
        # 'committedOffset == -1' means not exist last committed offset
        # startOffset = committedOffset + 1
        if committedOffset == -1:
            return -1
        else:
            return committedOffset + 1

    def inner_checkpoint(self):
        if self.consumerConfig.is_checkpoint_auto_commit():
            self.commit_offset(self.finishedOffset)

    def checkpoint(self, messageOffset=None):
        if not messageOffset:
            messageOffset = self.finishedOffset

        self.logger.info("start checkpoint: " + str(messageOffset))
        if self.consumerConfig.is_checkpoint_auto_commit():
            self.logger.info("You can not checkpoint through MessageCheckpointer when you set " +
                             "\"galaxy.talos.consumer.checkpoint.message.offset\" as \"true\"")
            return False

        if messageOffset <= self.lastCommitOffset or messageOffset > self.finishedOffset:
            self.logger.info("checkpoint messageOffset: " + str(messageOffset)
                             + " in wrong " + "range, lastCheckpoint messageOffset: "
                             + str(self.lastCommitOffset) + ", last " +
                             "deliver messageOffset: " + str(self.finishedOffset))
            return False

        try:
            self.commit_offset(messageOffset)
            return True
        except Exception as e:
            if isinstance(e, GalaxyTalosException):
                self.logger.error(e.errMsg)
            self.logger.error("Error when getting messages from topic: " +
                              self.topicAndPartition.topicTalosResourceName.topicTalosResourceName()
                              + " partition: " + str(self.topicAndPartition.partitionId)
                              + str(traceback.format_exc()))
            return False

    def commit_offset(self, messageOffset=None):
        checkPoint = CheckPoint(consumerGroup=self.consumerGroup,
                                topicAndPartition=self.topicAndPartition,
                                msgOffset=messageOffset,
                                workerId=self.workerId)
        # check whether to check last commit offset, firstCommit do not check
        if (not self.lastCommitOffset == -1) and self.consumerConfig.is_check_last_commit_offset():
            checkPoint.lastCommitOffset = self.lastCommitOffset

        updateOffsetRequest = UpdateOffsetRequest(checkPoint)
        updateOffsetResponse = self.consumerClient.update_offset(updateOffsetRequest)
        # update startOffset as next message
        if updateOffsetResponse.success:
            self.lastCommitOffset = messageOffset
            self.lastCommitTime = Utils.current_time_mills()
            self.logger.info("Worker: " + self.workerId + " commit offset: " +
                             str(messageOffset) + " for partition: " +
                             str(self.topicAndPartition.partitionId))
        else:
            self.logger.warn("Worker: " + self.workerId + " commit offset: " +
                             str(messageOffset) + " for partition: " +
                             str(self.topicAndPartition.partitionId) + " failed")

    def check_and_commit(self, isContinuous=None):
        if self.should_commit(isContinuous):
            try:
                self.inner_checkpoint()
            except Exception as e:
                if isinstance(e, GalaxyTalosException):
                    self.logger.error(e.errMsg)
                # when commitOffset failed, we just do nothing;
                self.logger.error("commit offset error, we skip to it"
                                  + str(traceback.format_exc()))

    def check_point(self):
        return self.check_point_with_offset(self.finishedOffset)

    def check_point_with_offset(self, commitOffset=None):
        self.logger.info("start checkpoint: " + str(commitOffset))
        if self.consumerConfig.is_checkpoint_auto_commit():
            self.logger.info("You can not checkpoint through MessageCheckpointer when you set " +
                             "\"galaxy.talos.consumer.checkpoint.message.offset\" as \"true\"")
            return False

        if commitOffset <= self.lastCommitOffset or commitOffset > self.finishedOffset:
            self.logger.info("checkpoint messageOffset: " + str(commitOffset) + " in wrong " +
                             "range, lastCheckpoint messageOffset: " + str(self.lastCommitOffset)
                             + ", last deliver messageOffset: " + str(self.finishedOffset))
            return False

        try:
            self.commit_offset(commitOffset)
            return True
        except Exception as e:
            if isinstance(e, GalaxyTalosException):
                self.logger.error(e.errMsg)
            self.logger.error("Error when getting messages from topic: " +
                              self.topicAndPartition.topicTalosResourceName.topicTalosResourceName
                              + " partition: " + str(self.topicAndPartition.partitionId)
                              + str(traceback.format_exc()))
            return False

